#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <netdb.h>
#include <errno.h>
#include <syslog.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <net/if.h>
#include <net/route.h>
#include <time.h>
#include <sys/time.h>
#include "csyncd.h" 

struct sockaddr_in sin_temp, sin_local, sin_remote, sin_reset;
int s,sr;

extern int errno;
extern char **environ;
int passive_up = 0; /* csyncd running flag */
int passive_present = 1; /* card inserted flag */
int active_up = 0; /* csyncd running flag */

/*
 * Log a string to syslog (and console if in debug mode)
 */
void logstr(int pri,char *str)
{
	time_t t;
	char buf[32];
	
	time(&t);
	strncpy(buf,ctime(&t),sizeof(buf));
	buf[24]=0;
#ifdef DEBUG
	printf("[%s] %s\n",buf,str);
#else
	syslog(LOG_DAEMON|pri,"%s",str);
#endif
}

/*
 * Figure out if we are the active or passive controller. Determined via a 
 * hardware read. For now we use conditional compilation. Sets the
 * passive_present flag as well.
 */
int determine_mode(void)
{
#ifdef ACTIVE
	logstr(LOG_INFO,"Entering active mode");
	return MODE_ACTIVE;
#else
	logstr(LOG_INFO,"Entering passive mode");
	return MODE_PASSIVE;
#endif
}

/*
 * Generate and return a valid IP address as a 32-bit integer
 */
unsigned int gen_ip_addr(void)
{
    unsigned a=0,b,c,d,ip;

    srand(time(NULL));
    do {
	a = 1+(int)(192.0*rand()/(RAND_MAX+1.0)); /* max of 192 here */
	b = 1+(int)(254.0*rand()/(RAND_MAX+1.0));
	c = 1+(int)(254.0*rand()/(RAND_MAX+1.0));
	d = 1+(int)(252.0*rand()/(RAND_MAX+1.0)); /* leave room to add 1 */
    } while(a<=150); /* guarantee a class C in 150.x.x.x to 192.x.x.x */
    ip = (a<<24)+(b<<16)+(c<<8)+d;

#ifdef DEBUG
    printf("gen_ip_addr: generated %d.%d.%d.%d (0x%X)\n",a,b,c,d,ip);
#endif
    return ip;
}

/*
 * Get the eth1 address and verify that "ip" does not fall into that network
 */
int same_net(unsigned ip)
{
    struct ifreq ifr;
    unsigned addr, mask;
    int sock;
    char buf[256];
    
    strcpy(ifr.ifr_name,"eth1");
    if((sock=socket(AF_INET,SOCK_DGRAM,0)) < 0) {
    	sprintf(buf,"check_eth1_addr: socket: %s",sys_errlist[errno]);
    	logstr(LOG_CRIT,buf);
    	exit(1);
    }
    if(ioctl(sock,SIOCGIFADDR,&ifr) < 0) {
    	sprintf(buf,"same_net: SIOCGIFADDR: %s",sys_errlist[errno]);
    	logstr(LOG_CRIT,buf);
    	exit(1);
    }
    memcpy(&addr,&ifr.ifr_addr.sa_data[2],sizeof(addr));
    if(ioctl(sock,SIOCGIFNETMASK,&ifr) < 0) {
    	sprintf(buf,"same_net: SIOCGIFNETMASK: %s",sys_errlist[errno]);
    	logstr(LOG_CRIT,buf);
    	exit(1);
    }
    memcpy(&mask,&ifr.ifr_netmask.sa_data[2],sizeof(mask));
    close(sock);
#ifdef DEBUG
    printf("same_net: eth1 %x/%x = %x ip %x/%x = %x\n",addr,mask,(addr & mask),
    ip,mask,(ip & mask));
#endif
    if((ip & mask) == (addr & mask)) /* same network? */
    	return 1;
    else
    	return 0;
}

/*
 * Loop until we see the config broadcast, then fill in the info struct
 */
void get_ipcfg_info(ipcfg_data_t *cfg)
{
    struct ifreq ifr;
    int sock,len;
    char buf[256];
    syncpkt_t pkt;
    
    logstr(LOG_INFO,"Configuring eth0 for addr 0.0.0.0, promisc mode");
    if((sock=socket(AF_INET,SOCK_DGRAM,0)) < 0) {
        sprintf(buf,"get_ipcfg_info: socket: %s",sys_errlist[errno]);
        logstr(LOG_CRIT,buf);
        exit(1);
    }
    strcpy(ifr.ifr_name,"eth0");
    ifr.ifr_addr.sa_family = AF_INET;
    memset(&ifr.ifr_addr.sa_data[2],0,sizeof(int));
    if(ioctl(sock,SIOCSIFADDR,&ifr) < 0) {
        sprintf(buf,"get_ipcfg_info: SIOCSIFADDR: %s",sys_errlist[errno]);
        logstr(LOG_CRIT,buf);
        exit(1);
    }
    strcpy(ifr.ifr_name,"eth0");
    ifr.ifr_flags = IFF_UP|IFF_BROADCAST|IFF_RUNNING|IFF_MULTICAST|IFF_PROMISC;
    if(ioctl(sock,SIOCSIFFLAGS,&ifr) < 0) {
        sprintf(buf,"get_ipcfg_info: SIOCSIFFLAGS: %s",sys_errlist[errno]);
        logstr(LOG_CRIT,buf);
        exit(1);
    }
    len = sizeof(sin_temp);
    memset(&sin_temp,0,len);
    sin_temp.sin_family = AF_INET;
    sin_temp.sin_addr.s_addr = INADDR_ANY;
    sin_temp.sin_port = htons(R_PORT);
    if( bind(sock,(struct sockaddr *)&sin_temp,len) < 0
     || getsockname(sock,(struct sockaddr *)&sin_temp, &len) < 0)
    {
	close(sock);
	sprintf(buf,"get_ipcfg_info: bind: %s",sys_errlist[errno]);
	logstr(LOG_CRIT,buf);
	exit(1);
    }
    logstr(LOG_INFO,"Waiting for BOOTSTRAP from active controller");
    recvfrom(sock,&pkt,sizeof(pkt),0,(struct sockaddr *)&sin_temp,&len);
    memcpy(cfg,&pkt.msg,sizeof(ipcfg_data_t));
#ifdef DEBUG
    printf("got some data: eth0_ip %x, eth1_ip %x, eth1_mask %x, eth1_bcast %x\n\n",
    ntohl(cfg->eth0_ip),ntohl(cfg->eth1_ip),ntohl(cfg->eth1_mask),ntohl(cfg->eth1_bcast));
#endif
    logstr(LOG_INFO,"Configuring local network from BOOTSTRAP packet");
    close(sock);
    return;
}

/*
 * Send config info to the other node and get it up and running
 */
int boot_passive(unsigned local,unsigned remote,unsigned mask,unsigned bcast)
{
    int len,sock;
    syncpkt_t pkt;
    ipcfg_data_t cfg;
    char buf[256];
    int val;
    
    len = sizeof(sin_temp);
    memset(&sin_temp,0,len);
    sin_temp.sin_family = AF_INET;
    sin_temp.sin_addr.s_addr = htonl((ntohl(local) & 0xffffff00)+255);
    sin_temp.sin_port = htons(R_PORT);
    pkt.datatype =  D_INVALID;
    pkt.oper = OP_BOOTSTRAP;
    cfg.eth0_ip = local;
    cfg.eth1_ip = remote;
    cfg.eth1_mask = mask;
    cfg.eth1_bcast = bcast;

    memcpy(pkt.msg,&cfg,sizeof(cfg));
    if((sock=socket(AF_INET,SOCK_DGRAM,0)) < 0) {
    	sprintf(buf,"boot_passive: socket: %s",sys_errlist[errno]);
    	logstr(LOG_CRIT,buf);
    	exit(1);
    }
    val = 1; /* enable broadcast perms on socket */
    if(setsockopt(sock,SOL_SOCKET,SO_BROADCAST,&val,sizeof(val)) < 0) {
    	sprintf(buf,"boot_passive: setsockopt: %s",sys_errlist[errno]);
    	logstr(LOG_CRIT,buf);
    	exit(1);
    }
    sendto(sock,&pkt,sizeof(pkt),0,(struct sockaddr *)&sin_temp,len);
    close(sock);
    return 0;
}

/* Configure eth0 with some IP and mask, broadcast to passive controller with
 * configure request. Listen for configure ack. Once the passive controller
 * acks, set up the sockets and return OK. If the passive controller times out
 * before acking, log an error<?> and sleep. Try again... Returns the socket
 */
int start_network(int mode)
{
    int len;
    unsigned local_addr,remote_addr,ipaddr,mask,bcast;
    u_short local_port,remote_port;
    char buf[256];
    struct ifreq ifr;
    struct rtentry rte;
    int sock;

    /* Use determine_mode() to see if we are active. If so, configure eth0
     * with a 'random' IP/mask and broadcast config request. Wait for passive
     * to ack then continue with socket stuff. If passive, wait for config
     * broadcast. Config eth0 from broadcast, then continue with sockets.
     */
    if(mode == MODE_ACTIVE) {
        logstr(LOG_INFO,"Configuring local network");
        do {
	    local_addr = htonl(gen_ip_addr());
	} while(same_net(local_addr));
    	remote_addr = htonl(ntohl(local_addr)+1);
    	local_port = L_PORT;
    	remote_port = R_PORT;

    	/* Get a socket */
    	if((sock=socket(AF_INET,SOCK_DGRAM,0)) < 0) {
    	    sprintf(buf,"start_network: socket: %s",sys_errlist[errno]);
    	    logstr(LOG_CRIT,buf);
    	    exit(1);
    	}

    	/* Config eth0 to our new addr */
    	strcpy(ifr.ifr_name,"eth0");
    	ifr.ifr_addr.sa_family = AF_INET;
    	memcpy(&ifr.ifr_addr.sa_data[2],&local_addr,sizeof(local_addr));
    	if(ioctl(sock,SIOCSIFADDR,&ifr) < 0) {
    	    sprintf(buf,"start_network: SIOCSIFADDR: %s",sys_errlist[errno]);
    	    logstr(LOG_CRIT,buf);
    	    exit(1);
    	}
    	mask = htonl(0xffffff00); /* we know it's class C.. */
    	strcpy(ifr.ifr_name,"eth0");
    	ifr.ifr_netmask.sa_family = AF_INET;
    	memcpy(&ifr.ifr_netmask.sa_data[2],&mask,sizeof(mask));
    	if(ioctl(sock,SIOCSIFNETMASK,&ifr) < 0) {
    	    sprintf(buf,"start_network: SIOCSIFNETMASK: %s",sys_errlist[errno]);
    	    logstr(LOG_CRIT,buf);
    	    exit(1);
    	}
	/* figure out the broadcast addr. again, we know it's class C */
	mask = htonl((ntohl(local_addr) & 0xffffff00)+255);
    	strcpy(ifr.ifr_name,"eth0");
    	ifr.ifr_broadaddr.sa_family = AF_INET;
    	memcpy(&ifr.ifr_broadaddr.sa_data[2],&mask,sizeof(mask));
    	if(ioctl(sock,SIOCSIFBRDADDR,&ifr) < 0) {
    	    sprintf(buf,"start_network: SIOCSIFBRDADDR: %s",sys_errlist[errno]);
    	    logstr(LOG_CRIT,buf);
    	    exit(1);
    	}
    	strcpy(ifr.ifr_name,"eth0");
    	ifr.ifr_flags = IFF_UP|IFF_BROADCAST|IFF_RUNNING|IFF_MULTICAST;
    	if(ioctl(sock,SIOCSIFFLAGS,&ifr) < 0) {
    	    sprintf(buf,"start_network: SIOCSIFFLAGS: %s",sys_errlist[errno]);
    	    logstr(LOG_CRIT,buf);
    	    exit(1);
    	}
    	/* now publish an entry for this net in the route table */
    	rte.rt_dev = "eth0";
    	rte.rt_dst.sa_family = AF_INET;
    	mask = htonl(ntohl(local_addr) & 0xffffff00);
    	memcpy(&rte.rt_dst.sa_data[2],&mask,sizeof(mask));
    	rte.rt_gateway.sa_family = AF_INET;
    	mask = 0;
    	memcpy(&rte.rt_gateway.sa_data[2],&mask,sizeof(mask));
    	rte.rt_flags = RTF_UP;
    	rte.rt_metric = 1; /* kernel subs 1 from this - BSD compat */
    	rte.rt_genmask.sa_family = AF_INET;
    	mask = htonl(0xffffff00);
    	memcpy(&rte.rt_genmask.sa_data[2],&mask,sizeof(mask));
    	rte.rt_mtu = 1500;
    	rte.rt_window = 0;
    	rte.rt_irtt = 0;
    	if(ioctl(sock,SIOCADDRT,&rte) < 0) {
    	    sprintf(buf,"start_network: SIOCADDRT: %s",sys_errlist[errno]);
    	    logstr(LOG_CRIT,buf);
    	    exit(1);
    	}

    	/* Get the eth1 info and send it to the passive node */
    	strcpy(ifr.ifr_name,"eth1");
    	if(ioctl(sock,SIOCGIFADDR,&ifr) < 0) {
    	    sprintf(buf,"start_network: SIOCGIFADDR: %s",sys_errlist[errno]);
    	    logstr(LOG_CRIT,buf);
    	    exit(1);
    	}
    	memcpy(&ipaddr,&ifr.ifr_addr.sa_data[2],sizeof(ipaddr));
    	strcpy(ifr.ifr_name,"eth1");
    	if(ioctl(sock,SIOCGIFNETMASK,&ifr) < 0) {
    	    sprintf(buf,"start_network: SIOCGIFNETMASK: %s",sys_errlist[errno]);
    	    logstr(LOG_CRIT,buf);
    	    exit(1);
    	}
    	memcpy(&mask,&ifr.ifr_netmask.sa_data[2],sizeof(mask));
    	strcpy(ifr.ifr_name,"eth1");
    	if(ioctl(sock,SIOCGIFBRDADDR,&ifr) < 0) {
    	    sprintf(buf,"start_network: SIOCGIFBRDADDR: %s",sys_errlist[errno]);
    	    logstr(LOG_CRIT,buf);
    	    exit(1);
    	}
    	memcpy(&bcast,&ifr.ifr_broadaddr.sa_data[2],sizeof(bcast));

    	close(sock);
    } else {
        ipcfg_data_t cfg;
        get_ipcfg_info(&cfg); /* get the local addr */
        local_addr = cfg.eth0_ip; /* already in net order */
        remote_addr = htonl(ntohl(local_addr) - 1);
    	local_port = R_PORT;
    	remote_port = L_PORT;

    	/* Get a socket */
    	if((sock=socket(AF_INET,SOCK_DGRAM,0)) < 0) {
    	    sprintf(buf,"start_network: socket: %s",sys_errlist[errno]);
    	    logstr(LOG_CRIT,buf);
    	    exit(1);
    	}

    	/* Config eth0 to our new addr */
    	strcpy(ifr.ifr_name,"eth0");
    	ifr.ifr_addr.sa_family = AF_INET;
    	memcpy(&ifr.ifr_addr.sa_data[2],&local_addr,sizeof(local_addr));
    	if(ioctl(sock,SIOCSIFADDR,&ifr) < 0) {
    	    sprintf(buf,"start_network: SIOCSIFADDR: %s",sys_errlist[errno]);
    	    logstr(LOG_CRIT,buf);
    	    exit(1);
    	}
    	mask = htonl(0xffffff00); /* we know it's class C.. */
    	strcpy(ifr.ifr_name,"eth0");
    	ifr.ifr_netmask.sa_family = AF_INET;
    	memcpy(&ifr.ifr_netmask.sa_data[2],&mask,sizeof(mask));
    	if(ioctl(sock,SIOCSIFNETMASK,&ifr) < 0) {
    	    sprintf(buf,"start_network: SIOCSIFNETMASK: %s",sys_errlist[errno]);
    	    logstr(LOG_CRIT,buf);
    	    exit(1);
    	}
	/* figure out the broadcast addr. again, we know it's class C */
	mask = htonl((ntohl(local_addr) & 0xffffff00)+255);
    	strcpy(ifr.ifr_name,"eth0");
    	ifr.ifr_broadaddr.sa_family = AF_INET;
    	memcpy(&ifr.ifr_broadaddr.sa_data[2],&mask,sizeof(mask));
    	if(ioctl(sock,SIOCSIFBRDADDR,&ifr) < 0) {
    	    sprintf(buf,"start_network: SIOCSIFBRDADDR: %s",sys_errlist[errno]);
    	    logstr(LOG_CRIT,buf);
    	    exit(1);
    	}
    	strcpy(ifr.ifr_name,"eth0");
    	ifr.ifr_flags = IFF_UP|IFF_BROADCAST|IFF_RUNNING|IFF_MULTICAST;
    	if(ioctl(sock,SIOCSIFFLAGS,&ifr) < 0) {
    	    sprintf(buf,"start_network: SIOCSIFFLAGS: %s",sys_errlist[errno]);
    	    logstr(LOG_CRIT,buf);
    	    exit(1);
    	}
    	/* now publish an entry for this net in the route table */
    	rte.rt_dev = "eth0";
    	rte.rt_dst.sa_family = AF_INET;
    	mask = htonl(ntohl(local_addr) & 0xffffff00);
    	memcpy(&rte.rt_dst.sa_data[2],&mask,sizeof(mask));
    	rte.rt_gateway.sa_family = AF_INET;
    	mask = 0;
    	memcpy(&rte.rt_gateway.sa_data[2],&mask,sizeof(mask));
    	rte.rt_flags = RTF_UP;
    	rte.rt_metric = 1; /* kernel subs 1 from this - BSD compat */
    	rte.rt_genmask.sa_family = AF_INET;
    	mask = htonl(0xffffff00);
    	memcpy(&rte.rt_genmask.sa_data[2],&mask,sizeof(mask));
    	rte.rt_mtu = 1500;
    	rte.rt_window = 0;
    	rte.rt_irtt = 0;
    	if(ioctl(sock,SIOCADDRT,&rte) < 0) {
    	    sprintf(buf,"start_network: SIOCADDRT: %s",sys_errlist[errno]);
    	    logstr(LOG_CRIT,buf);
    	    exit(1);
    	}
    	/* Config eth1 from config packet but mark it down */
    	strcpy(ifr.ifr_name,"eth1");
    	ifr.ifr_addr.sa_family = AF_INET;
    	memcpy(&ifr.ifr_addr.sa_data[2],&cfg.eth1_ip,sizeof(cfg.eth1_ip));
    	if(ioctl(sock,SIOCSIFADDR,&ifr) < 0) {
    	    sprintf(buf,"start_network: SIOCSIFADDR: %s",sys_errlist[errno]);
    	    logstr(LOG_CRIT,buf);
    	    exit(1);
    	}
    	strcpy(ifr.ifr_name,"eth1");
    	ifr.ifr_netmask.sa_family = AF_INET;
    	memcpy(&ifr.ifr_netmask.sa_data[2],&cfg.eth1_mask,sizeof(cfg.eth1_mask));
    	if(ioctl(sock,SIOCSIFNETMASK,&ifr) < 0) {
    	    sprintf(buf,"start_network: SIOCSIFNETMASK: %s",sys_errlist[errno]);
    	    logstr(LOG_CRIT,buf);
    	    exit(1);
    	}
    	strcpy(ifr.ifr_name,"eth1");
    	ifr.ifr_broadaddr.sa_family = AF_INET;
    	memcpy(&ifr.ifr_broadaddr.sa_data[2],&cfg.eth1_bcast,sizeof(cfg.eth1_bcast));
    	if(ioctl(sock,SIOCSIFBRDADDR,&ifr) < 0) {
    	    sprintf(buf,"start_network: SIOCSIFBRDADDR: %s",sys_errlist[errno]);
    	    logstr(LOG_CRIT,buf);
    	    exit(1);
    	}
    	strcpy(ifr.ifr_name,"eth1");
    	ifr.ifr_flags = IFF_BROADCAST|IFF_MULTICAST; /* if is down */
    	if(ioctl(sock,SIOCSIFFLAGS,&ifr) < 0) {
    	    sprintf(buf,"start_network: SIOCSIFFLAGS: %s",sys_errlist[errno]);
    	    logstr(LOG_CRIT,buf);
    	    exit(1);
    	}
	close(sock);
    }

    sprintf(buf,"Loc addr: %08x/%d Rem addr: %08x/%d",ntohl(local_addr),
    local_port,ntohl(remote_addr),remote_port);
    logstr(LOG_INFO,buf);
    
    /* Set up the sockets */
    s = socket(AF_INET, SOCK_DGRAM, 0);
    if(s < 0)
    {
        sprintf(buf,"socket: %s",sys_errlist[errno]);
        logstr(LOG_CRIT,buf);
	exit(1);
    }

    len = sizeof(sin_local);
    memset(&sin_local,0,len);
    sin_local.sin_family = AF_INET;
    sin_local.sin_addr.s_addr = local_addr;
    sin_local.sin_port = htons(local_port);
    if( bind(s,(struct sockaddr *)&sin_local,len) < 0
     || getsockname(s,(struct sockaddr *)&sin_local, &len) < 0)
    {
	close(s);
	sprintf(buf,"bind: %s",sys_errlist[errno]);
	logstr(LOG_CRIT,buf);
	exit(1);
    }

    len = sizeof(sin_remote);
    memset(&sin_remote,0,len);
    sin_remote.sin_family = AF_INET;
    sin_remote.sin_addr.s_addr = remote_addr;
    sin_remote.sin_port = htons(remote_port);

    if(mode == MODE_PASSIVE) { /* bind a port to listen for reset requests */
        sr = socket(AF_INET, SOCK_DGRAM, 0);
        if(sr < 0)
        {
            sprintf(buf,"socket: %s",sys_errlist[errno]);
            logstr(LOG_CRIT,buf);
	    exit(1);
        }
        len = sizeof(sin_reset);
        memset(&sin_reset,0,len);
        sin_reset.sin_family = AF_INET;
        sin_reset.sin_addr.s_addr = local_addr;
        sin_reset.sin_port = htons(RESET_PORT);
        if( bind(sr,(struct sockaddr *)&sin_reset,len) < 0
          || getsockname(sr,(struct sockaddr *)&sin_reset, &len) < 0)
        {
	    close(sr);
	    sprintf(buf,"bind: %s",sys_errlist[errno]);
	    logstr(LOG_CRIT,buf);
	    exit(1);
        }
    } else {
        len = sizeof(sin_reset);
        memset(&sin_reset,0,len);
        sin_reset.sin_family = AF_INET;
        sin_reset.sin_addr.s_addr = remote_addr;
        sin_reset.sin_port = htons(RESET_PORT);

	/* Get passive up - must be done after active socket setup */
        logstr(LOG_INFO,"Configuring the passive controller...");
	while(passive_up != 1) {
    	    boot_passive(remote_addr,ipaddr,mask,bcast);
    	    if(get_reply() == OP_ACK)
    	        passive_up = 1;
    	    else
    	        sleep(STARTNET_RETRY_WAIT);
    	}
    	logstr(LOG_INFO,"Passive controller just came up");
    }
    return s;
}

/*
 * Reset the passive controller and active too if needed
 */
void do_reset(char **argv)
{
	syncpkt_t pkt;
	static int reset_count = 0;
	struct ifreq ifr;
	ipcfg_data_t cfg;
	char buf[256];
	int sock;
	
	passive_up = 0; /* flag controller down */
       	logstr(LOG_CRIT,"Passive controller not responding - resetting.");
      	pkt.datatype = D_INVALID;
       	pkt.oper = OP_RECONFIG;
       	memset(pkt.msg,0,sizeof(pkt.msg));
       	sendto(s,&pkt,sizeof(pkt),0,(struct sockaddr *)&sin_reset,sizeof(sin_reset));
	sleep(5); /* give it time to start */

        if((sock=socket(AF_INET,SOCK_DGRAM,0)) < 0) {
            sprintf(buf,"do_reset: socket: %s",sys_errlist[errno]);
            logstr(LOG_CRIT,buf);
            exit(1);
        }
        
  	/* Send a BOOTSTRAP packet to reconfigure the passive node */
        strcpy(ifr.ifr_name,"eth1");
        if(ioctl(sock,SIOCGIFADDR,&ifr) < 0) {
            sprintf(buf,"do_reset: SIOCGIFADDR: %s",sys_errlist[errno]);
            logstr(LOG_CRIT,buf);
            exit(1);
        }
        memcpy(&cfg.eth1_ip,&ifr.ifr_addr.sa_data[2],sizeof(cfg.eth1_ip));
        strcpy(ifr.ifr_name,"eth1");
        if(ioctl(sock,SIOCGIFNETMASK,&ifr) < 0) {
            sprintf(buf,"do_reset: SIOCGIFNETMASK: %s",sys_errlist[errno]);
            logstr(LOG_CRIT,buf);
            exit(1);
        }
        memcpy(&cfg.eth1_mask,&ifr.ifr_netmask.sa_data[2],sizeof(cfg.eth1_mask));
        strcpy(ifr.ifr_name,"eth1");
        if(ioctl(sock,SIOCGIFBRDADDR,&ifr) < 0) {
            sprintf(buf,"do_reset: SIOCGIFBRDADDR: %s",sys_errlist[errno]);
            logstr(LOG_CRIT,buf);
            exit(1);
        }
        memcpy(&cfg.eth1_bcast,&ifr.ifr_broadaddr.sa_data[2],sizeof(cfg.eth1_bcast));

	/* Send the packet */
        logstr(LOG_INFO,"Configuring the passive controller...");
	while(passive_up != 1) {
            boot_passive(sin_local.sin_addr.s_addr,cfg.eth1_ip,cfg.eth1_mask,cfg.eth1_bcast);
            if(get_reply() == OP_ACK) {
                passive_up = 1;
                reset_count = 0;
            } else {
                passive_up = 0;
                reset_count++;
		logstr(LOG_ERR,"Restart failed. Passive failed to ACK.");
	       	if(reset_count > 3) {
       	  	    logstr(LOG_CRIT,"Too many passive resets. Active restarting.");
       	  	    close(sock);
       	    	    close(s);
       	    	    reset_count=0;
       	    	    execve(argv[0],argv,environ);
       	    	    logstr(LOG_CRIT,"Active restart failed! Can't continue!");
       	    	    exit(1);
       		}
       		sleep(5); /* try again in 5 sec */
            }
	}
	close(sock);
        logstr(LOG_INFO,"Passive controller just came up");
  	return;
}

#define DO_UPDATE(x) \
	if((x) > 0) { \
	    if(++p_fail_count > 3) { \
                p_fail_count = 0; \
                do_reset(argv); \
            } \
	}
/*
 * Daemonize and start the action
 */
int main(int argc, char **argv)
{
    int mode;
    struct timeval tv;
    fd_set rfds;
    int ret;
    char buf[128];
    syncpkt_t pkt;
    int salen;
    int p_fail_count = 0; /* "update failed" counter - should probably age
    			      this and zero it after an hour w/ no failures */
    
    mode = determine_mode();
    s = start_network(mode);
    if(mode == MODE_ACTIVE) {

        while(1) {
            /* if card inserted but csynd isn't running, loop until it is */
            while(passive_up != 1 && passive_present) do_reset(argv);
	    DO_UPDATE(update_arp_table());
	    DO_UPDATE(update_route_table());
	    DO_UPDATE(update_ipfwd_table());
	    DO_UPDATE(update_ipautofw_table());
	    sleep(SYNC_TIME);
        }
    } else { /* mode == MODE_PASSIVE */
        /* signal that we're here */
        send_ack(D_INVALID);
        while(1) {
            /* Is there a packet waiting? */
            tv.tv_sec = 0;
            tv.tv_usec = 5000; /* Wait up to 1/2 sec */
            FD_ZERO(&rfds);
            FD_SET(s,&rfds);
            if((ret=select(s+1,&rfds,0,0,&tv)) > 0)
                receive_packet();
            else if(ret < 0) {
                sprintf(buf,"select: %s",sys_errlist[errno]);
            	logstr(LOG_ERR,buf);
            } else {
                /* Check hardware here to see if the active controller is
                   alive or not. Switch state if it isn't */
            }
            /* Check the reset port to see if we should restart */
            tv.tv_sec = 0;
            tv.tv_usec = 2500; /* Wait 1/4 sec for reset */
            FD_ZERO(&rfds);
            FD_SET(sr,&rfds);
            if((ret=select(sr+1,&rfds,0,0,&tv)) > 0) {
                salen = sizeof(sin_reset);
                recvfrom(sr,&pkt,sizeof(pkt),0,(struct sockaddr *)&sin_reset,&salen);
                if(pkt.datatype == D_INVALID && pkt.oper == OP_RECONFIG) {
                    logstr(LOG_CRIT,"RECONFIG received... restarting!");
                    close(s);
                    close(sr);
            	    execve(argv[0],argv,environ);
            	    logstr(LOG_CRIT,"RECONFIG request received but restart failed!");
            	    send_nak(D_INVALID); /* indicate failure to active */
            	} else
            	    logstr(LOG_WARNING,"Invalid data received on reset port!");
            } else if(ret < 0) {
            	sprintf(buf,"select: %s",sys_errlist[errno]);
            	logstr(LOG_CRIT,buf);
            	/* maybe try to restart here instead of exiting? */
            	exit(1);
            }
        }
    }
    exit(0);
}
