/* File: config.c
 * Desc: Code for dealing with memory based configuration data..
 * Auth: San Mehat
 * Note:
 */
#include <stdio.h>
#include <config.h>
#include <fcntl.h>
#include <string.h>
#include <netinet/in.h>
#include <time.h>

#include <syslog.h>

// Global data
char Stored[ITEM_MAX][FIELD_MAX];

// Prototypes
int checkConfigChanged();

char *mytime()
{
	time_t t;
	struct tm *tm_ptr;
	static char buf[32];
	
	time(&t);
	tm_ptr = localtime(&t);
	strftime(buf,sizeof(buf),"%c",tm_ptr);
	return buf;
}

/*
 * Load configuration data into 'Tconfigp_ptr'.
 * return 0 if the config data was loaded successfully and -1 otherwise
 */
int config_load(CONFIGURATION *Tconfigp_ptr)
{
	FILE 			*Tfilep_ptr;
	int 			I_rc;
	char 			SZ_line[512];
	char 			*bword;
	char 			SZ_copy[512];
	int 			I_line=0;
	CONFIG_PARTITION	*Tpartition_new=0;
	FILE			*Tfilep_sumpipe;
	int  			I_prisum = 0,I_secsum = 0;
	char			SZ_filename[255];
	char			SZ_tmp[255];

	sprintf(SZ_tmp,"/usr/bin/sum %s", ZAPHOD_CONFIG_FILE1);
	if (!(Tfilep_sumpipe=popen(SZ_tmp, "r")))
	{
		perror("popen");
		return -1;
	}
	fscanf(Tfilep_sumpipe,"%d",&I_prisum);
	pclose(Tfilep_sumpipe);

	sprintf(SZ_tmp,"/usr/bin/sum %s", ZAPHOD_CONFIG_FILE2);
	if (!(Tfilep_sumpipe=popen(SZ_tmp, "r")))
	{
		perror("popen");
		return -1;
	}
	fscanf(Tfilep_sumpipe,"%d",&I_secsum);
	pclose(Tfilep_sumpipe);

	if (I_prisum!=I_secsum)
	{
		printf("[%s] Warning: Zaphod configuration inconsistancy (pri=%d, sec=%d)\n",
			mytime(),
			I_prisum,
			I_secsum);
		printf("[%s] Using backup configuration data\n",mytime());
		strcpy(SZ_filename,ZAPHOD_CONFIG_FILE2);
	}
	else
	{
		printf("[%s] Configuration data integrity verified\n",mytime());
		strcpy(SZ_filename,ZAPHOD_CONFIG_FILE1);
	}

	if (!(Tfilep_ptr=fopen(SZ_filename,"r")))
	{
		perror("fopen");
		printf("[%s] Unable to open %s\n",mytime(),SZ_filename);
		return -1;
	}

	while (fgets(SZ_line,sizeof(SZ_line),Tfilep_ptr))
	{
		I_line++;
		SZ_line[strlen(SZ_line)-1]=0;
//		printf("[config_load] Processing <%s>\n",SZ_line);
		strcpy(SZ_copy,SZ_line);
		if ((SZ_line[0]=='#')||(SZ_line[0]=='\n')||(SZ_line[0]==0))
			continue;
		if (!(bword=strtok(SZ_line,"=")))
			goto syntax_error;
		if (!strcmp(bword,"SERVERNAME"))
		{
			if (Tpartition_new)
				goto yes_partition;
			if (!(bword=strtok(NULL,"=")))
				goto bad_addr;
			strcpy(Tconfigp_ptr->SZ_servername,bword);
		}
		else if (!strcmp(bword,"IPADDR_EXT"))
		{
			if (Tpartition_new)
				goto yes_partition;
			if (!(bword=strtok(NULL,"=")))
				goto bad_addr;
			if ((Tconfigp_ptr->UI_ext_ipaddr=inet_addr(bword))<0)
				goto bad_addr;
		}
		else if (!strcmp(bword,"NETMASK_EXT"))
		{
			if (Tpartition_new)
				goto yes_partition;
			if (!(bword=strtok(NULL,"=")))
				goto bad_addr;
			if ((Tconfigp_ptr->UI_ext_netmask=inet_addr(bword))<0)
				goto bad_addr;
		}
		else if (!strcmp(bword,"IPADDR_INT"))
		{
			if (Tpartition_new)
				goto yes_partition;
			if (!(bword=strtok(NULL,"=")))
				goto bad_addr;
			if ((Tconfigp_ptr->UI_int_ipaddr=inet_addr(bword))<0)
				goto bad_addr;
		}
		else if (!strcmp(bword,"NETMASK_INT"))
		{
			if (Tpartition_new)
				goto yes_partition;
			if (!(bword=strtok(NULL,"=")))
				goto bad_addr;
			if ((Tconfigp_ptr->UI_int_netmask=inet_addr(bword))<0)
				goto bad_addr;
		}
		else if (!strcmp(bword,"GATEWAY"))
		{
			if (Tpartition_new)
				goto yes_partition;
			if (!(bword=strtok(NULL,"=")))
				goto bad_addr;
			if ((Tconfigp_ptr->UI_gateway=inet_addr(bword))<0)
				goto bad_addr;
		}
		else if (!strcmp(bword,"NAMESERVER"))
		{
			if (Tpartition_new)
				goto yes_partition;
			if (!(bword=strtok(NULL,"=")))
				goto bad_addr;
			if ((Tconfigp_ptr->UI_nameserver=inet_addr(bword))<0)
				goto bad_addr;
		}
		else if (!strcmp(bword,"RO"))
		{
			int I_c1;
			char SZ_dest[32];
			char SZ_gw[32];
			char SZ_type[5];
			struct sockaddr_in *sock;
			struct sockaddr_in *socktmp;
			struct rtentry *routep;

			if (!Tpartition_new)
			{
				/* Find first empty route position in the global routes list */
				for (I_c1=0;Tconfigp_ptr->Srtentry_route[I_c1].rt_metric!=0;I_c1++);
				if (I_c1==6)
				{
					printf("[config_load] Routing table entries exhausted\n");
					goto out;
				}
				routep=&Tconfigp_ptr->Srtentry_route[I_c1];
			}
			else
			{
				routep=&Tpartition_new->Srtentry_route;
			}
			if (!(bword=strtok(NULL,":")))
				goto bad_addr;
			strcpy(SZ_dest,bword);
			if (!(bword=strtok(NULL,":")))
				goto bad_addr;
			strcpy(SZ_gw,bword);
			if (!(bword=strtok(NULL,":")))
				goto syntax_error;
			strcpy(SZ_type,bword);
			if (SZ_type[0]=='h')
				routep->rt_flags=(RTF_UP | RTF_HOST);
			else if (SZ_type[0]=='n')
				routep->rt_flags=(RTF_UP | RTF_GATEWAY);
			else
			{
				printf("[config_load] Routing entry type invalid (%s)\n",SZ_type);
				goto out;
			}

			/* Fill in the route destination */
			sock = (struct sockaddr_in *) &routep->rt_dst;
			sock->sin_family=AF_INET;
			sock->sin_port=0;
			if ((sock->sin_addr.s_addr=inet_addr(SZ_dest))<0)
				goto bad_addr;

			/* Now the route gateway */
			sock = (struct sockaddr_in *) &routep->rt_gateway;
			sock->sin_family=AF_INET;
			sock->sin_port=0;
			if ((sock->sin_addr.s_addr=inet_addr(SZ_gw))<0)
				goto bad_addr;

			/* Now the route netmask/genmask thingy */
			sock = (struct sockaddr_in *) &routep->rt_genmask;
			sock->sin_family=AF_INET;
			sock->sin_port=0;
			socktmp= (struct sockaddr_in *) &Tconfigp_ptr->Srtentry_route[I_c1].rt_dst;
			if (socktmp->sin_addr.s_addr==0)
				sock->sin_addr.s_addr=0;
			else
			if (!(sock->sin_addr.s_addr=Tconfigp_ptr->UI_int_netmask))
			{
				printf("[config_load] Syntax error on line %d\n",I_line);
				printf("[config_load] RO (route) specified before ZM (ZIB netmask)\n");
				goto out;
			}
			routep->rt_metric=1;
		}
		else if (!strcmp(bword,"CE"))
		{
			if (Tpartition_new)
				goto yes_partition;
			if (!(bword=strtok(NULL,":")))
				goto syntax_error;
			Tconfigp_ptr->UC_ctrletheraddr[0]= (char) strtol(bword,(char **)NULL,16);
			if (!(bword=strtok(NULL,":")))
				goto syntax_error;
			Tconfigp_ptr->UC_ctrletheraddr[1]= (char) strtol(bword,(char **)NULL,16);
			if (!(bword=strtok(NULL,":")))
				goto syntax_error;
			Tconfigp_ptr->UC_ctrletheraddr[2]= (char) strtol(bword,(char **)NULL,16);
			if (!(bword=strtok(NULL,":")))
				goto syntax_error;
			Tconfigp_ptr->UC_ctrletheraddr[3]= (char) strtol(bword,(char **)NULL,16);
			if (!(bword=strtok(NULL,":")))
				goto syntax_error;
			Tconfigp_ptr->UC_ctrletheraddr[4]= (char) strtol(bword,(char **)NULL,16);
			if (!(bword=strtok(NULL,":")))
				goto syntax_error;
			Tconfigp_ptr->UC_ctrletheraddr[5]= (char) strtol(bword,(char **)NULL,16);
		}
		else if (!strcmp(bword,"SYNC_FREQ"))
		{
/*
			if (Tpartition_new)
				goto yes_partition;
			if (!(bword=strtok(NULL," ")))
				goto syntax_error;
			if (!(Tconfigp_ptr->UI_syncmngrscanfreq=atoi(bword)))
				goto syntax_error;
*/
		}
		else if (!strcmp(bword,"PING_FREQ"))
		{
			if (Tpartition_new)
				goto yes_partition;
			if (!(bword=strtok(NULL,":")))
				goto syntax_error;
			if (!(Tconfigp_ptr->UI_cardsvcsscanfreq=atoi(bword)))
				goto syntax_error;
		}
		else if (!strcmp(bword,"SL"))
		{
/*
			if (Tpartition_new)
				goto yes_partition;
			if (!(bword=strtok(NULL,":")))
				goto syntax_error;
			if (!(Tconfigp_ptr->UC_slavememthrottle=(char)atoi(bword)))
				goto syntax_error;
*/
		}
		else if (!strcmp(bword,"LOADBAL_GLOB"))
		{
			if (Tpartition_new)
				goto yes_partition;
			if (!(bword=strtok(NULL,":")))
				goto syntax_error;
			strcpy(Tconfigp_ptr->SZ_globalloadbalancer,bword);
		}	
		else if (!strcmp(bword,"ServerAdmin"))
		{
			if (Tpartition_new)
				goto yes_partition;
			if ((bword=strtok(NULL,"*")) == NULL)
				strcpy(Tconfigp_ptr->SZ_email, "");
			else
			{
				strcpy(Tconfigp_ptr->SZ_email, bword);
			}
			// Email addresses should not contain '*'s
			if ((bword = strtok(NULL,"*")) != NULL)
				goto syntax_error;
		}	
		else if (!strcmp(bword,"TZ"))
		{
			if (Tpartition_new)
				goto yes_partition;
			if (!(bword=strtok(NULL,"*"))) // '*' should never appear in a filepath
				goto syntax_error;
			strcpy(Tconfigp_ptr->SZ_timezone,bword);
		}			
		else if (!strcmp(bword,"NS"))
		{
			if (Tpartition_new)
				goto yes_partition;
			if (!(bword=strtok(NULL,"=")))
				goto bad_addr;
			if ((Tconfigp_ptr->UI_nfsserver=inet_addr(bword))<0)
				goto bad_addr;
		}
		else if (!strcmp(bword,"NB"))
		{
			if (Tpartition_new)
				goto yes_partition;
			if (!(bword=strtok(NULL,"*"))) // '*' should never appear in a filepath
				goto syntax_error;
			strcpy(Tconfigp_ptr->SZ_binarypath,bword);
		}			
		else if (!strcmp(bword,"VHOST_BEGIN"))
		{
			if (Tpartition_new)
			{
				printf("[config_load] Syntax Error on line %d of configuration file\n",I_line);
				printf("[config_load] PB specified when allready in partition\n");
				goto out;
			}
			else
			{
				/* New partition definition */
				if (!Tconfigp_ptr->Tpartition_root)
				{
					/* First partition */
					Tpartition_new= (CONFIG_PARTITION *) malloc (sizeof(CONFIG_PARTITION));
					memset(Tpartition_new,0,sizeof(CONFIG_PARTITION));
					Tconfigp_ptr->Tpartition_root=Tpartition_new;
				}
				else
				{
					CONFIG_PARTITION* Tpartitionp_ptr;
					/* Additional partition */
					for (Tpartitionp_ptr=Tconfigp_ptr->Tpartition_root;
						Tpartitionp_ptr->next;
						Tpartitionp_ptr=Tpartitionp_ptr->next);
					Tpartition_new= (CONFIG_PARTITION *) malloc (sizeof(CONFIG_PARTITION));
					memset(Tpartition_new,0,sizeof(CONFIG_PARTITION));
					Tpartitionp_ptr->next = Tpartition_new;
				}
			/* Partition structure linked in */
			}
		}
		else if (!strcmp(bword,"IPADDR"))
		{
			if (!Tpartition_new)
				goto no_partition;
			if (!(bword=strtok(NULL,"=")))
				goto bad_addr;
			if ((Tpartition_new->UI_aliasaddr=inet_addr(bword))<0)
				goto bad_addr;
		}
		else if (!strcmp(bword,"NETMASK"))
		{
			if (!Tpartition_new)
				goto no_partition;
			if (!(bword=strtok(NULL,"=")))
				goto bad_addr;
			if ((Tpartition_new->UI_aliasmask=inet_addr(bword))<0)
				goto bad_addr;
		}
		else if (!strcmp(bword,"PORT"))
		{
			if (!Tpartition_new)
				goto no_partition;
			if (!(bword=strtok(NULL,":")))
				goto syntax_error;
			if (!(Tpartition_new->US_portbinding=htons((unsigned short)atoi(bword))))
				goto syntax_error;
		}
		else if (!strcmp(bword,"Options"))
		{
			if (!Tpartition_new)
				goto no_partition;
			if (!(bword=strtok(NULL,":")))
				goto syntax_error;
			if (!(Tpartition_new->C_optionflags=(char)atoi(bword)))
				goto syntax_error;
		}
		else if (!strcmp(bword,"ServerName"))
		{
			if (!Tpartition_new)
				goto no_partition;
			if (!(bword=strtok(NULL,"*"))) /* * should never appear in name */
				goto syntax_error;
			strcpy(Tpartition_new->SZ_servername,bword);
		}
		else if (!strcmp(bword,"DocumentRoot"))
		{
			if (!Tpartition_new)
				goto no_partition;
			if (!(bword=strtok(NULL,"*"))) /* * should never appear in a filepath */
				goto syntax_error;
			strcpy(Tpartition_new->SZ_documentpath,bword);
		}
		else if (!strcmp(bword,"ScriptAlias"))
		{
			if (!Tpartition_new)
				goto no_partition;
			if (!(bword=strtok(NULL,"*"))) /* * should never appear in a filepath */
				goto syntax_error;
			strcpy(Tpartition_new->SZ_cgipath,bword);
		}
		else if (!strcmp(bword,"SSLEnable"))
		{
			if (!Tpartition_new)
				goto no_partition;
			if (!(bword=strtok(NULL,":")))
				goto syntax_error;
			Tpartition_new->C_ssl = atoi(bword);
		}
		else if (!strcmp(bword,"SSLCertificate"))
		{
			if (!Tpartition_new)
				goto no_partition;
			bword = strtok(NULL, "*");
			if (bword == NULL)
				strcpy(Tpartition_new->SZ_sslcertificate, "");
			else
			{
				strcpy(Tpartition_new->SZ_sslcertificate, bword);
				// '*' should never appear in a filepath
				if ((bword=strtok(NULL,"*")) != NULL)
					goto syntax_error;
			}
		}
		else if (!strcmp(bword,"LOADBAL"))
		{
			if (!Tpartition_new)
				goto no_partition;
			if (!(bword=strtok(NULL,"*"))) /* * should never appear in a filepath */
				goto syntax_error;
			strcpy(Tpartition_new->SZ_loadbalancer,bword);
		}
		else if (!strcmp(bword,"NO"))
		{
			CONFIG_NODELIST *Tnodelistp_new;
			CONFIG_NODELIST *Tnodelistp_walk;

			if (!Tpartition_new)
				goto no_partition;

			if (!Tpartition_new->Tnodelist_root)
			{
				Tnodelistp_new = (CONFIG_NODELIST *) malloc(sizeof(CONFIG_NODELIST));
				memset(Tnodelistp_new,0,sizeof(CONFIG_NODELIST));
				Tpartition_new->Tnodelist_root=Tnodelistp_new;
			}
			else
			{
				for (Tnodelistp_walk=Tpartition_new->Tnodelist_root;
					Tnodelistp_walk->next;
					Tnodelistp_walk=Tnodelistp_walk->next);
				Tnodelistp_new= (CONFIG_NODELIST *) malloc (sizeof(CONFIG_NODELIST));
				memset(Tnodelistp_new,0,sizeof(CONFIG_NODELIST));
				Tnodelistp_walk->next=Tnodelistp_new;
			}

			/* Created and linked... now populate the sucker */
			if (!(bword=strtok(NULL,":")))
				goto syntax_error;
			if (!(Tnodelistp_new->I_nodenumber=(unsigned int)atoi(bword)))
				goto syntax_error;

			if (!(bword=strtok(NULL,":")))
				goto syntax_error;
			if (!(Tnodelistp_new->UI_throughputchoke=(unsigned int)atoi(bword)))
				goto syntax_error;

			if (!(bword=strtok(NULL,":")))
				goto syntax_error;
			if (!(Tnodelistp_new->UC_priority=(unsigned char)atoi(bword)))
				goto syntax_error;
		}
		else if (!strcmp(bword,"VHOST_END"))
		{
			if (!Tpartition_new)
			{
				printf("[config_load] Syntax Error on line %d of configuration file\n",I_line);
				printf("[config_load] PE specified when not in partition\n");
				goto out;
			}
			Tpartition_new=0;
		}
		else
		{
			goto syntax_error;
		}
	}

	fclose(Tfilep_ptr);

	return(0);

syntax_error:
	printf("[config_load] Syntax Error on line %d of configuration file\n",I_line);
	printf("[config_load] (%s)\n",SZ_copy);
	fclose(Tfilep_ptr);
	return(-1);

bad_addr:
	printf("[config_load] Bad inet address on line %d of configuration file\n",I_line);
	printf("[config_load] (%s)\n",SZ_copy);
	fclose(Tfilep_ptr);
	return(-1);
out:
	fclose(Tfilep_ptr);
	return(-1);
no_partition: 
	printf("[config_load] Syntax Error on line %d of configuration file\n",I_line);
	printf("[config_load] Partition token detected not within partition\n");
	fclose(Tfilep_ptr);
	return(-1);
yes_partition: 
	printf("[config_load] Syntax Error on line %d of configuration file\n",I_line);
	printf("[config_load] Non-Partition token detected within partition\n");
	fclose(Tfilep_ptr);
	return(-1);
}

/*
 * Free memory associated with partitions (ie virtual hosts) stored in the
 * specified configuration structure.
 */
void destroy_partitions(CONFIGURATION *Tconfigp_ptr)
{
	CONFIG_PARTITION *Tpartitionp_begin=Tconfigp_ptr->Tpartition_root;
	CONFIG_PARTITION *Tpartitionp_ptr;
	CONFIG_NODELIST  *Tnodelistp_begin;
	CONFIG_NODELIST  *Tnodelistp_ptr;

	// While there are partitions in this config
	while (Tpartitionp_begin)
	{
		Tnodelistp_begin=Tpartitionp_begin->Tnodelist_root;
		
		// While there are nodes in the nodelist
		while(Tnodelistp_begin)
		{
			Tpartitionp_begin->Tnodelist_root=Tnodelistp_begin->next; /* Short circuit */
			free (Tnodelistp_begin);
			Tnodelistp_begin=Tpartitionp_begin->Tnodelist_root; /* reset */
		}
		Tconfigp_ptr->Tpartition_root = Tpartitionp_begin->next; /* short circuit */
		free (Tpartitionp_begin);
		Tpartitionp_begin=Tconfigp_ptr->Tpartition_root;
	}

}

/*
 * Save the configuration data stored in 'Tconfigp_ptr' to the configuration
 * files defined in config.h.
 * return -1 if there was an error
 * return 0 if neither system nor web server config data changed
 * return 1 if system data changed
 * return 2 if system data did not change and web server data changed
 */
int config_save(CONFIGURATION* Tconfigp_ptr)
{
	int	I_c1, ret = 0;
	char	buf[32];
	char	SZ_tmp[256];
	char	SZ_tmpfilename[128];
	FILE	*Tfilep_ptr;

	// build name of temp file to which settings will be saved before
	// overwriting original config files
	sprintf(SZ_tmpfilename, "%s.new", ZAPHOD_SYSCONFIG_FILE);

	// open our config file
	if (!(Tfilep_ptr=fopen(SZ_tmpfilename ,"w")))
		return(-1);

	// Controller Global Data
	fprintf( Tfilep_ptr, "SERVERNAME=%s.%s\n", Stored[HOSTNAME], Stored[DOMAIN] );

	fprintf( Tfilep_ptr, "IPADDR_EXT=%s\n", itod(dtoi(Stored[EXTERNAL_IPADDR])) );
	fprintf( Tfilep_ptr, "NETMASK_EXT=%s\n", itod(dtoi(Stored[EXTERNAL_NETMASK])) );

	fprintf( Tfilep_ptr, "IPADDR_INT=%s\n", itod(dtoi(Stored[INTERNAL_IPADDR])) );
	fprintf( Tfilep_ptr, "NETMASK_INT=%s\n", itod(dtoi(Stored[INTERNAL_NETMASK])) );

	fprintf( Tfilep_ptr, "GATEWAY=%s\n", itod(dtoi(Stored[GATEWAY])) );
	fprintf( Tfilep_ptr, "NAMESERVER=%s\n", itod(dtoi(Stored[NAMESERVER])) );

	// Loop for routes
	for (I_c1=0;Tconfigp_ptr->Srtentry_route[I_c1].rt_metric!=0 && I_c1<6 ; I_c1++)
	{
		char SZ_type[5];
		char SZ_gw[32];
		char SZ_dest[32];
		struct sockaddr_in *dst;
		struct sockaddr_in *gw;

		if (Tconfigp_ptr->Srtentry_route[I_c1].rt_flags==(RTF_UP|RTF_HOST))
			strcpy( SZ_type, "h" );
		else if (Tconfigp_ptr->Srtentry_route[I_c1].rt_flags==(RTF_UP|RTF_GATEWAY))
			strcpy( SZ_type, "n" );
		else
		{
			// Error encountered
			return(-1);
		}
		dst = (struct sockaddr_in *) &Tconfigp_ptr->Srtentry_route[I_c1].rt_dst;
		strcpy( SZ_dest, itod(dst->sin_addr.s_addr) );
		gw = (struct sockaddr_in *) &Tconfigp_ptr->Srtentry_route[I_c1].rt_gateway;
		strcpy( SZ_gw, itod(htonl(gw->sin_addr.s_addr)) );
		fprintf( Tfilep_ptr, "RO=%s:%s:%s\n", SZ_dest, SZ_gw, SZ_type );
	}

	// Mac Address
	sprintf( buf, "%x:%x:%x:%x:%x:%x", \
		Tconfigp_ptr->UC_ctrletheraddr[0], \
		Tconfigp_ptr->UC_ctrletheraddr[1], \
		Tconfigp_ptr->UC_ctrletheraddr[2], \
		Tconfigp_ptr->UC_ctrletheraddr[3], \
		Tconfigp_ptr->UC_ctrletheraddr[4], \
		Tconfigp_ptr->UC_ctrletheraddr[5] ); 
	fprintf( Tfilep_ptr, "CE=%s\n", buf );
	
	fprintf( Tfilep_ptr, "SYNC_FREQ=%d\n", Tconfigp_ptr->UI_syncmngrscanfreq );
	fprintf( Tfilep_ptr, "PING_FREQ=%d\n", Tconfigp_ptr->UI_cardsvcsscanfreq );
	fprintf( Tfilep_ptr, "SL=%d\n", Tconfigp_ptr->UC_slavememthrottle );
	
	fprintf( Tfilep_ptr, "LOADBAL_GLOB=%s\n", Stored[LOADBAL]);

	fprintf( Tfilep_ptr, "ServerAdmin=%s\n", Stored[SERVERADMIN]);

	fprintf( Tfilep_ptr, "TZ=%s\n", Stored[TIMEZONE] );
	fprintf( Tfilep_ptr, "NS=%s\n", itod(dtoi(Stored[NFSSERV])) );
	fprintf( Tfilep_ptr, "NB=%s\n", Stored[NFSBIN] );

	// Partition Data (Virtual Servers)
	if ( Tconfigp_ptr->Tpartition_root )
	{
		// We have partitions defined
		CONFIG_PARTITION *Tpartitionp_ptr;
		int I_index;
		for ( I_index = 0, Tpartitionp_ptr = Tconfigp_ptr->Tpartition_root; \
			Tpartitionp_ptr; \
			I_index++, Tpartitionp_ptr = Tpartitionp_ptr->next )
		{
			char SZ_type[5];
			char SZ_gw[32];
			char SZ_dest[32];

			struct sockaddr_in *dst;
			struct sockaddr_in *gw;

			// Begin Partition
			fprintf( Tfilep_ptr, "VHOST_BEGIN=%d\n",I_index );

			fprintf( Tfilep_ptr, "IPADDR=%s\n",itod(htonl(Tpartitionp_ptr->UI_aliasaddr)));
			fprintf( Tfilep_ptr, "NETMASK=%s\n",itod(htonl(Tpartitionp_ptr->UI_aliasmask)));

			// More fun with routes!
			if ( Tpartitionp_ptr->Srtentry_route.rt_metric == 1 )
			{
				if (Tpartitionp_ptr->Srtentry_route.rt_flags==(RTF_UP|RTF_HOST))
					strcpy( SZ_type, "h" );
				else if (Tpartitionp_ptr->Srtentry_route.rt_flags==(RTF_UP|RTF_GATEWAY))
					strcpy( SZ_type, "n" );
				else
					return(-1); // We encountered an error
				dst = (struct sockaddr_in *) &Tpartitionp_ptr->Srtentry_route.rt_dst;
				strcpy( SZ_dest, itod(htonl(dst->sin_addr.s_addr)) );
				gw = (struct sockaddr_in *) &Tpartitionp_ptr->Srtentry_route.rt_gateway;
				strcpy( SZ_gw, itod(htonl(gw->sin_addr.s_addr)) );
			
				fprintf( Tfilep_ptr, "RO=%s:%s:%s\n", SZ_dest, SZ_gw, SZ_type );
			}

			fprintf( Tfilep_ptr, "ServerName=%s\n", Tpartitionp_ptr->SZ_servername );
			fprintf( Tfilep_ptr, "PORT=%d\n", htons(Tpartitionp_ptr->US_portbinding));
			fprintf( Tfilep_ptr, "DocumentRoot=%s\n", Tpartitionp_ptr->SZ_documentpath );
			if (strlen(Tpartitionp_ptr->SZ_cgipath) > 0)
				fprintf( Tfilep_ptr, "ScriptAlias=%s\n", Tpartitionp_ptr->SZ_cgipath );
			fprintf( Tfilep_ptr, "LOADBAL=%s\n", Tpartitionp_ptr->SZ_loadbalancer);
			fprintf( Tfilep_ptr, "Options=%d\n", Tpartitionp_ptr->C_optionflags );
			fprintf( Tfilep_ptr, "SSLEnable=%d\n", Tpartitionp_ptr->C_ssl );
			fprintf( Tfilep_ptr, "SSLCertificate=%s\n", Tpartitionp_ptr->SZ_sslcertificate );

			// Nodelists
			if ( Tpartitionp_ptr->Tnodelist_root )
			{
				CONFIG_NODELIST *Tnodelistp_walk;
				for ( Tnodelistp_walk=Tpartitionp_ptr->Tnodelist_root; \
					Tnodelistp_walk; \
					Tnodelistp_walk=Tnodelistp_walk->next )
				{
					fprintf( Tfilep_ptr, "NO=%d:%d:%d\n", \
						Tnodelistp_walk->I_nodenumber, \
						Tnodelistp_walk->UI_throughputchoke, \
						Tnodelistp_walk->UC_priority );
				}
			}

			// End Partition
			fprintf( Tfilep_ptr, "VHOST_END=%d\n", I_index );
		}
	}

	// close our config file
	fclose(Tfilep_ptr);

	// Save three copies of the config file. Two are identical and the third
	// does not have any virtual host info. The third is used by card services
	// which is responsible for storing this info to the EEPROM.

	// 1.	Copy the temp file to config file 1
	sprintf(SZ_tmp, "cp %s %s", SZ_tmpfilename, ZAPHOD_CONFIG_FILE1);
	switch (system(SZ_tmp))
	{
	    case -1:
	    case 127:
		return -1;
	}
	
	// 2.	Copy the first 16 lines (the system config data) to the system
	//	config file.
	sprintf(SZ_tmp, "head -n 16 %s > %s 2> /dev/null",
		SZ_tmpfilename, ZAPHOD_SYSCONFIG_FILE);
	switch (system(SZ_tmp))
	{
	    case -1:
	    case 127:
		return -1;
	}
	
	// Before writing the backup config file, check changes between the
	// new and old configuration settings. If there is an error in doing
	// assume that all settings (system and web server) changed so that
	// zaphod will be restarted in which case all settings will be applied.
	ret = checkConfigChanged();
	if (ret == -1)
	    ret = 1;

	// 3.	Rename the temp file to be the backup config file.
	sprintf(SZ_tmp, "mv %s %s", SZ_tmpfilename, ZAPHOD_CONFIG_FILE2);
	switch (system(SZ_tmp))
	{
	    case -1:
	    case 127:
		return -1;
	}

	// 4.	Now save apache formatted config data
	if (apache_config_save(Tconfigp_ptr) < 0)
	{
	    return -1;
	}
	else
	{
	    return ret;
	}
}

/*
 * Save apache settings in apache config file format.
 * return -1 if there is an error and 0 otherwise
 */
int apache_config_save(CONFIGURATION* Tconfigp_ptr)
{
	int	I_c1;
	char	buf[32];
	char	SZ_tmpfilename[128];
	char	SZ_tmp[256];
	FILE	*Tfilep_ptr;

	// Make a copy of the apache defaults
	system("cp " APACHE_DEFAULTS_FILE " " ZAPHOD_APACHE_FILE);	

	// Open this file to which we will append apache-formatted config
	// data	supplied through this instance of serial config.
	if (!(Tfilep_ptr=fopen(ZAPHOD_APACHE_FILE, "a")))
		return(-1);

	// Mark the start of zaphodcfg data
	fprintf( Tfilep_ptr, "\n\n# START OF ZAPHODCFG SETTINGS\n\n");
	
	// Controller Global Data
//	fprintf( Tfilep_ptr, "SERVERNAME=%s.%s\n", Stored[HOSTNAME], Stored[DOMAIN] );
//	fprintf( Tfilep_ptr, "SL=%d\n", Tconfigp_ptr->UC_slavememthrottle );	
	fprintf( Tfilep_ptr, "ServerAdmin %s\n\n", Stored[SERVERADMIN]);
	fprintf( Tfilep_ptr, "ServerRoot %s\n\n", Stored[NFSBIN] );

	fprintf( Tfilep_ptr, "NameVirtualHost %s\n\n",
		itod(dtoi(Stored[EXTERNAL_IPADDR])));

	// Partition Data (Virtual Servers)
	if ( Tconfigp_ptr->Tpartition_root )
	{
		// We have partitions defined
		
		int			I_SSLFlag = 0;
		CONFIG_PARTITION	*Tpartitionp_ptr;
		
		Tpartitionp_ptr = Tconfigp_ptr->Tpartition_root;
		while (Tpartitionp_ptr != NULL)
		{
			char SZ_type[5];
			char SZ_gw[32];
			char SZ_dest[32];

			struct sockaddr_in *dst;
			struct sockaddr_in *gw;

			// Begin Partition
			fprintf(Tfilep_ptr, "<VirtualHost %s%s>\n",
				itod(dtoi(Stored[EXTERNAL_IPADDR])),
				I_SSLFlag ? ":443" : "");

//			fprintf( Tfilep_ptr, "AI=%s\n",itod(htonl(Tpartitionp_ptr->UI_aliasaddr)));
//			fprintf( Tfilep_ptr, "AM=%s\n",itod(htonl(Tpartitionp_ptr->UI_aliasmask)));

			fprintf( Tfilep_ptr, "ServerName %s\n", Tpartitionp_ptr->SZ_servername );
			fprintf( Tfilep_ptr, "DocumentRoot %s\n", Tpartitionp_ptr->SZ_documentpath );
			if (strlen(Tpartitionp_ptr->SZ_cgipath) > 0)
				fprintf( Tfilep_ptr, "ScriptAlias /cgi-bin/ %s\n", Tpartitionp_ptr->SZ_cgipath );
			switch (Tpartitionp_ptr->C_optionflags)
			{
				case OPT_INCLUDE_NONE:
					fprintf(Tfilep_ptr, "Options -Includes\n");
					break;

				case OPT_INCLUDE_NOEXEC:
					fprintf(Tfilep_ptr, "Options IncludesNOEXEC\n");
					break;

				case OPT_INCLUDE_FULL:
					fprintf(Tfilep_ptr, "Options Includes\n");
					break;
			}			

			if (Tpartitionp_ptr->C_ssl)
			{
				fprintf(Tfilep_ptr, "SSLEnable\n");
				fprintf(Tfilep_ptr, "SSLCertificateFile %s\n",
					Tpartitionp_ptr->SZ_sslcertificate);
				fprintf(Tfilep_ptr, "SSLCACertificateFile %s\n",
					Tpartitionp_ptr->SZ_sslcertificate);

				I_SSLFlag = (I_SSLFlag == 0) ? 1 : 0;
			}
			else
			{
				fprintf(Tfilep_ptr, "SSLDisable\n");
			}

			// End Partition
			fprintf( Tfilep_ptr, "</VirtualHost>\n\n" );

			// If this virtual host supports SSL, we will add another
			// entry for port 443 => do not advance to the next virtual
			// host just yet			
			if (!I_SSLFlag)
				Tpartitionp_ptr = Tpartitionp_ptr->next;
		}
	}

	// close our config file
	fclose(Tfilep_ptr);

	// Save a backup of the existing apache config file
	sprintf(SZ_tmpfilename, "%s/conf/%s", Stored[NFSBIN], APACHE_CONFIG_FILE);
	sprintf(SZ_tmp, "cp %s %s.backup", SZ_tmpfilename, SZ_tmpfilename);
	system(SZ_tmp);

	// Replace the existing zaphod apache config file with the one that
	// we just created.
	sprintf(SZ_tmpfilename, "%s/conf/%s", Stored[NFSBIN], APACHE_CONFIG_FILE);
	sprintf(SZ_tmp, "cp %s %s", ZAPHOD_APACHE_FILE, SZ_tmpfilename);
	switch (system(SZ_tmp))
	{
	    case -1:
	    case 127:
		return -1;

	    default:
		return 0;
	}
}

char *itod (int i)
{
	// 32-bit int to dotted IP notation
	static char buf[16];
	int a,b,c,d;
	d = i & 0xff;
	i = i >> 8;
	c = i & 0xff;
	i = i >> 8;
	b = i & 0xff;
	i = i >> 8;
	a = i & 0xff;

	sprintf ( buf, "%d.%d.%d.%d", a,b,c,d );
	return ( (char *)buf );
}

unsigned int dtoi(char *str)
{
	// dotted IP to unsigned int
	int a,b,c,d;
	int ip;
	sscanf(str,"%d.%d.%d.%d",&a,&b,&c,&d);
	ip=(a<<24)+(b<<16)+(c<<8)+d;
	return ip;
}

void serverToHostAndDomain(char* SZ_server,char* SZ_host, char* SZ_domain)
{
	char* SZ_temp, *SZ_temp1;

	SZ_temp = (char*)malloc(strlen(SZ_server) + 1);
	if (SZ_temp != NULL)
	{
		strcpy(SZ_temp, SZ_server);
		SZ_temp1 = strtok(SZ_temp,".");
		if (SZ_temp1 != NULL)
		{
			strcpy(SZ_host, SZ_temp1);
			if (strlen(SZ_host) < strlen(SZ_server))
				strcpy(SZ_domain, &(SZ_server[strlen(SZ_host) + 1]));
			else
				strcpy(SZ_domain,"");
		}
		else
		{
			strcpy(SZ_host, "");
			strcpy(SZ_domain, "");
		}
		free(SZ_temp);
	}
}

/*
 * Return the process ID of the running CardServices process or -1 if it is
 * not running. This is based on the assumption that CardServices writes its
 * pid to a file on starting and removes the file when it has shutdown.
 */
int getCardServicesPID()
{
	FILE	*fileptr;
	int	pid;

	if (!(fileptr = fopen(CARD_SERVICES_PID_FILE,"r")))
	{
		return -1;
	}
	else
	{
		fscanf(fileptr, "%d", &pid);
		fclose(fileptr);
		return pid;
	}
}

/*
 * Check if configuration data changed since the last invokation.
 * NOTE: this assumes that a new config file has been written but its
 * backup has not yet been written.
 * return -1 if there was an error while checking
 * return 0	if there were no changes
 * return 1	if there were system config changes
 * return 2	if there were no system config changes and there were web
 *		server config changes.
 */
int checkConfigChanged()
{
    FILE	*Tfilep_pipe;
    int		line_num, ret = 0;
    char	SZ_tmp[512];

    // Use diff to compare the system config data with that of the old.
    // Filter all but line numbers so that we can determine if lines in the
    // system section (the first 16 lines) changed vs lines in the web server
    // section (> 16 lines).
    sprintf(SZ_tmp,"diff %s %s | grep \"^[^<^>^-]\"",
	    ZAPHOD_CONFIG_FILE2, ZAPHOD_SYSCONFIG_FILE);
    if (!(Tfilep_pipe=popen(SZ_tmp, "r")))
    {
	perror("popen");
	return -1;
    }
    else
    {
	if (fscanf(Tfilep_pipe, "%d", &line_num) == EOF)
	    ret = 0;	// no changes
	else
	{
	    if (line_num <= 16)
		ret = 1;	// sys config info changed
	    else
		ret = 2;	// web server config info changed
	}
	pclose(Tfilep_pipe);
	return ret;
    }
}
